from typing import Dict, List
import dataiku
import logging

def replace_tables_in_ast(ast: Dict, datasets: List):
    valid_aliases = set()

    def _collect_valid_aliases(node: Dict):
        assert(node["type"] == "TABLE")
        table_name = node["name"]

        if "alias" in node:
            valid_aliases.add(node["alias"])

    def _collect_aliases_rec_list(node: List):
        for item in node:
            if isinstance(item, dict):
                _collect_aliases_rec_dict(item)
            elif isinstance(item, list):
                _collect_aliases_rec_list(item)

    def _collect_aliases_rec_dict(node: Dict):
        for key in node.keys():
            if (key == "table" or key == "tableLike" or key == "from") and isinstance(node[key], dict) and "type" in node[key] and node[key]["type"] == "TABLE":
                _collect_valid_aliases(node[key])
            else:
                if isinstance(node[key], dict):
                    _collect_aliases_rec_dict(node[key])
                elif isinstance(node[key], list):
                    _collect_aliases_rec_list(node[key])

    def _expand_table_node(node: Dict):
        assert(node["type"] == "TABLE")
        table_name = node["name"]

        if not table_name in datasets:
            if not table_name in valid_aliases:
                raise Exception("Illegal table access: %s" % table_name)
            else:
                # Leave the alias alone
                return

        location_info = dataiku.Dataset(table_name).get_location_info()
        if location_info.get("locationInfoType") != "SQL":
            raise ValueError("Cannot only execute query on an SQL dataset")

        catalog_name = location_info.get("info").get("catalog")
        schema_name = location_info.get("info").get("schema")
        table_name = location_info.get("info").get("table")

        if catalog_name is not None:
            node["catalog"] = catalog_name
        if schema_name is not None:
            node["schema"] = schema_name
        if table_name is not None:
            node["name"] = table_name
        
    def _replace_tables_rec_list(node: List):
        for item in node:
            if isinstance(item, dict):
                _replace_tables_rec_dict(item)
            elif isinstance(item, list):
                _replace_tables_rec_list(item)

    def _replace_tables_rec_dict(node: Dict):
        for key in node.keys():
            if (key == "table" or key == "tableLike" or key == "from") and isinstance(node[key], dict) and "type" in node[key] and node[key]["type"] == "TABLE":
                _expand_table_node(node[key])
            else:
                if isinstance(node[key], dict):
                    _replace_tables_rec_dict(node[key])
                elif isinstance(node[key], list):
                    _replace_tables_rec_list(node[key])

    _collect_aliases_rec_dict(ast._query)
    logging.info("Collected valid aliases: %s" % valid_aliases)
    _replace_tables_rec_dict(ast._query)